# Available Options and Their Meaning

To select a call type for a Protobuf service method, you must specify one or more boolean options.
Gorums currently supports the following call types (these cannot be combined):

| Call type           | Gorums option        | Description                                                         |
| ------------------- | -------------------- | ------------------------------------------------------------------- |
| Ordered RPC         | no option            | FIFO-ordered synchronous RPC to a single node.                      |
| Unicast             | `gorums.unicast`     | FIFO-ordered one-way asynchronous unicast.                          |
| Multicast           | `gorums.multicast`   | FIFO-ordered one-way asynchronous multicast.                        |
| Quorum Call         | `gorums.quorumcall`  | FIFO-ordered synchronous quorum call on a configuration of nodes.   |
| Correctable         | `gorums.correctable` | Asynchronous quorum call with *incremental consistency guarantees*. |

All call types rely on a streaming RPC call named `NodeStream` to send and receive messages to/from individual nodes.
However, the server API generated by Gorums does not expose the gRPC stream-based API.
Instead, the generated API is similar to unary gRPC, except for the handling of return values is a bit different.

Each call type may also specify some advanced options:

| Name               | Gorums option               | Type   | Description                                                                                    |
| ------------------ | --------------------------- | ------ | ---------------------------------------------------------------------------------------------- |
| Custom return type | `gorums.custom_return_type` | string | Makes the quorum function return a different type than the RPC response type.                  |
| Per node arguments | `gorums.per_node_arg`       | bool   | Makes it possible to send different request objects to each node in a call.                    |
| Asynchronous       | `gorums.async`              | bool   | Makes the quorum call run asynchronously and returns a "promise" object to fetch the result.   |

The following matrix shows which call types can be combined with the different options:

| Option             | Ordered RPC | Unicast | Multicast | Quorum Call | Correctable |
| ------------------ | ----------- | ------- | --------- | ----------- | ----------- |
| Custom return type | N/A         | N/A     | N/A       | Yes         | Yes         |
| Per node arguments | N/A         | N/A     | Yes       | Yes         | Yes         |
| Asynchronous       | No          | N/A     | N/A       | Yes         | No          |

The correctable quorum call is an asynchronous call type that comes in two variants.
In the first variant, each node of the configuration sends only a single response.
In the second variant, each node may return multiple responses.
The second variant uses the gRPC server-side stream API and is indicated with the `stream` keyword associated with the return type.
See `ReadCorrectableStream` method below for an example.

Note that the `gorums.correctable` option should not be combined with the `gorums.async` option.

## Example Protobuf Definitions

Below is shown a list of RPC service methods, illustrating how to configure the different variants.

```proto
service Storage {
  // ReadOrdered is a FIFO-ordered RPC.
  rpc ReadOrdered(ReadRequest) returns (State) {}

  // ReadQC is a FIFO-ordered synchronous quorum call.
  rpc ReadQC(ReadRequest) returns (State) {
    option (gorums.quorumcall) = true;
  }

  // ReadAsync is an asynchronous quorum call that
  // returns a promise object for retrieving results.
  rpc ReadAsync(ReadRequest) returns (State) {
    option (gorums.async) = true;
    option (gorums.quorumcall) = true;
  }

  // ReadCustomReturn is a FIFO-ordered synchronous quorum call with a custom return type.
  rpc ReadCustomReturn(ReadRequest) returns (State) {
    option (gorums.quorumcall) = true;
    option (gorums.custom_return_type) = "MyState";
  }

  // ReadCorrectable is an asynchronous correctable quorum call that
  // returns a correctable object for retrieving results.
  rpc ReadCorrectable(ReadRequest) returns (State) {
    option (gorums.correctable) = true;
  }

  // ReadCorrectableStream is an asynchronous correctable quorum call that
  // returns a correctable object for retrieving results over a stream.
  rpc ReadCorrectableStream(ReadRequest) returns (stream State) {
    option (gorums.correctable) = true;
  }

  // WriteMulticast is an asynchronous multicast to all nodes in a configuration.
  // No replies are collected.
  rpc WriteMulticast(State) returns (Empty) {
    option (gorums.multicast) = true;
  }

  // WritePerNode is a synchronous quorum call, where,
  // for each node, a provided function is called to determine
  // the argument to be sent to that node.
  rpc WritePerNode(State) returns (WriteResponse) {
    option (gorums.quorumcall)  = true;
    option (gorums.per_node_arg)  = true;
  }
}
```
